#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/boot.h>

#include "types.h"

#define PROG_ACTIVE (!(PIND & (1 << 2)))

/* This jumps to address 0, the beginning of the application section.
 * This is called whenever an error occurs.
 */
void run_user_code()
{
	/* Put registers back to default values */
	TCCR1B = 0x00;
	TCNT1 = 0;
	UCSRB = 0x00;		/* Disable the UART before clearing TXC */
	UCSRA = 0x40;		/* Clear TXC by writing a one to it */
	/* UCSRC is already in the default configuration */
	GIFR = 0xff;
	TIFR = 0xff;
	DDRD = 0x00;
	PORTD = 0x00;

	/* Enable reading from RWW (the application section).
	 *
	 * This is automatically disabled for page writes.
	 */
	boot_rww_enable();

	/* Jump to address 0 */
	asm(	"eor r30, r30\n\t"
		"eor r31, r31\n\r"
		"ijmp\n\r");
}

/* Sends a character to the UART */
void ser_send(u8 ch)
{
	/* Wait for the transmit buffer to become empty
	 * (last byte has finished transmitting)
	 */
	while (!(UCSRA & (1 << UDRE)))
		;

	/* Send the character */
	UDR = ch;
}

/* Reads one character from the UART and returns it.
 *
 * If PROG goes inactive while waiting for a character, the user code is executed.
 */
u8 ser_receive()
{
	while (!(UCSRA & (1 << RXC)))
	{
		if (!PROG_ACTIVE)
			run_user_code();
	}

	return UDR;
}

/* Reads one character from the UART and returns it.
 *
 * Timer1 is reset before waiting. If it overflows before a character is received,
 * then the user code is executed.
 *
 * If PROG goes inactive while waiting for a character, the user code is executed.
 */
u8 ser_receive_timed()
{
	TCNT1 = 0;
	while (!(UCSRA & (1 << RXC)))
	{
		if (TIFR & (1 << TOV1) || !PROG_ACTIVE)
			run_user_code();
	}

	return UDR;
}

/* Receives a character and compares it to the given character.
 *
 * If the received character is different, took too long to arrive, or PROG dropped,
 * the user code is executed.
 */
void receive_handshake(u8 expect_ch)
{
	if (ser_receive_timed() != expect_ch)
		run_user_code();
}

/* Make the compiler omit the return instruction */
int main() __attribute__((__noreturn__));

int main()
{
	u16 page_base;
	u8 i;

	/* Disable interrupts in case we didn't cleanly enter this code */
	cli();

	/* If PROG is inactive, run user code */
	if (!PROG_ACTIVE)
		run_user_code();

	/* Set up the I/O ports */
	PORTD = 0xff;
	DDRD = 0xe2;

	/* Set up the UART */
	UBRRL = 7;
	UCSRB = 0x18;
	UCSRA = 0x40;
	UCSRC = 0x86;

	/* Start timer 1 for serial port timeout, clk/64 */
	TCCR1B = 0x02;

	/* Send handshake trigger to the host */
	ser_send('T');
	ser_send('1');

	/* Wait for handshake from the host */
	receive_handshake('H');
	receive_handshake(1);

	/* First address in the page to be loaded */
	page_base = 0;

	while (1)
	{
		u16 checksum = 0;
		u16 addr = page_base;

		/* Receive a page of data from the host */
		PORTD = 0x7f;
		for (i = 0; i < SPM_PAGESIZE; i += 2)
		{
			u16 word = ser_receive();
			word |= ser_receive() << 8;

			boot_page_fill(addr, word);
			addr += 2;

			checksum += word;
		}

		/* Erase the page */
		PORTD = 0xdf;
		boot_page_erase(page_base);
		boot_spm_busy_wait();

		/* Write the page */
		PORTD = 0xbf;
		boot_page_write(page_base);
		boot_spm_busy_wait();

		/* Advance to the next page */
		page_base += SPM_PAGESIZE;

		/* Send checksum to host */
		ser_send(checksum & 0xff);
		ser_send(checksum >> 8);

		/* Don't continue downloading code if the application section is full */
		if (page_base >= 0x1800)
			run_user_code();
	}
}

